#! /usr/bin/env python3
import json
import traceback
from argparse import ArgumentParser
from datetime import datetime
from shutil import move
from sys import exit, stdin
from OpenSSL import crypto

import dns.resolver
import email
import git
import glob
import os
import pipes
import re
import socket
import subprocess
import time
import urllib3
import typing
import filelock


# Default configuration
DEFAULT_BASEDIR = os.path.expanduser('~') + '/Certificates'

# CRL URLs (are stable for a given CA generation, so they identify the generation)
CRL_URLS_TO_GENERATION = {
    'http://cdp1.pca.dfn.de/kit-ca/pub/crl/cacrl.crl': 'KIT-CA-G1',
    'http://cdp1.pca.dfn.de/kit-ca-g2/pub/crl/cacrl.crl': 'KIT-CA-G2',
    'http://GEANT.crl.sectigo.com/GEANTPersonalECCCA4.crl': 'geant-tcs-sectigo',
    'http://GEANT.crl.sectigo.com/GEANTPersonalCA4.crl': 'geant-tcs-sectigo',
    'http://GEANT.crl.sectigo.com/GEANTOVRSACA4.crl': 'geant-tcs-sectigo',
    'http://GEANT.crl.sectigo.com/GEANTOVECCCA4.crl': 'geant-tcs-sectigo',
    'http://GEANT.crl.sectigo.com/GEANTEVRSACA4.crl': 'geant-tcs-sectigo',
    'http://GEANT.crl.sectigo.com/GEANTEVECCCA4.crl': 'geant-tcs-sectigo',
    'http://GEANT.crl.sectigo.com/GEANTeScienceSSLECCCA4.crl': 'geant-tcs-sectigo',
    'http://GEANT.crl.sectigo.com/GEANTeScienceSSLCA4.crl': 'geant-tcs-sectigo',
    'http://GEANT.crl.sectigo.com/GEANTeSciencePersonalECCCA4.crl': 'geant-tcs-sectigo',
    'http://GEANT.crl.sectigo.com/GEANTeSciencePersonalCA4.crl': 'geant-tcs-sectigo',
}

SIGNAL_URL = "https://automation.cert.kit.edu/signal/"

# Parse args
ACTIONS = [
    'add',
    'addmail',
    'autoexpire',
    'autoprocess',
    'autorevoke',
    'classify',
    'expire',
    'init',
    'processmail',
    'revoke',
    'revokemail',
    'testnotification'
]


LOCK_FILE = "/tmp/process_certificate.lock"
LOCK_TIMEOUT = 10 * 60  # In seconds

# Git commit message
commit_message = ''
commits = []
repo: typing.Optional[git.Repo] = None

notifier_config: typing.Optional[dict] = None


# Make directory
def _mkdir(new_dir):
    """works the way a good mkdir should :)
        - already exists, silently complete
        - regular file in the way, raise an exception
        - parent directory(ies) does not exist, make
          them as well
    """
    if os.path.isdir(new_dir):
        pass
    elif os.path.isfile(new_dir):
        raise OSError(f"A file with the same name as the desired dir, '{new_dir}', already exists.")
    else:
        head, tail = os.path.split(new_dir)
        if head and not os.path.isdir(head):
            _mkdir(head)
        if tail:
            os.mkdir(new_dir)


# Check if symlink exists
def _check_link(source, target):
    if os.path.exists(target):
        if os.path.islink(target):
            if os.readlink(target) == source:
                return True
            else:
                raise OSError(f"A symlink with the same name, '{target}', but a different target already exists.")
        else:
            raise OSError("a file with the same name, '{target}', already exists.")
    else:
        return False


# Create symlink
def _link_file(source, target):
    """Create a symlink from source to target;
        - If (correct) symlink already exists, silently complete
        - if symlink does not exist, do the symlink
        - otherwise, raise an exception
    """
    if _check_link(source, target):
        pass
    else:
        os.symlink(source, target)
        if args.git:
            repo.index.add([target])


# Move a file
def _movefile(file, target):
    if args.git:
        repo.index.move([file, target])
    else:
        move(file, target)


# Load a mail file
# Takes file name as input
# If file name is '-', then stdin is read instead
def read_mail(mail_input):
    if mail_input == '-':
        # Read mail from stdin
        mail_as_string = ''
        for line in stdin:
            mail_as_string += line
        result = email.message_from_string(mail_as_string)
    else:
        # Read mail file
        with open(mail_input, "r") as fh:
            result = email.message_from_file(fh)

    return result


def set_symlink_timestamps_to_target_timestamps(symlink, reference):
    if os.path.islink(symlink):
        os.system(f"touch --no-dereference --reference {reference} {symlink}")


# Process an X.509 certificate
# Takes a PEM-formatted string as input
def add_certificate(cert_string):
    # Parse the certificate
    certificate = crypto.load_certificate(crypto.FILETYPE_PEM, cert_string)

    # Get the certificate serial number and assemble the archive file name
    # The DFN uses the format 'cert-$SERIAL.pem', so that is what we will
    # use as well
    archived_cert_filename = f"cert-{certificate.get_serial_number()}.pem"

    # Generate a pretty filename.  For the time being, we will use the
    # format
    #  $SANITIZED_CN-$EXPIRY_DATE-$SERIAL.pem
    # with
    #  $SANITIZED_CN being the CN, but sanitized for filename compatibility
    #   (blanks replaced by underscores)
    #  $EXPIRY_DATE being the expiration date(!, not the timestamp) in
    #   ISO8601 format (YYYY-MM-DD)
    #  $SERIAL being the serial number.

    # Clean CN
    clean_cn = certificate.get_subject().commonName
    if not clean_cn:
        clean_cn = certificate.get_subject().emailAddress.replace("@", "_at_")

    for bad_char in [' ', '/', ',', '\'', '*', ':', '\\', '(', ')']:
        clean_cn = clean_cn.replace(bad_char, '_')

    # Convert expiry date
    expiry_date = time.strftime('%Y-%m-%d', time.strptime(certificate.get_notAfter().decode("utf-8"), '%Y%m%d%H%M%SZ'))

    # Construct filename
    pretty_filename = f"{clean_cn}-{expiry_date}-{certificate.get_serial_number()}.pem"

    # Write the actual certificate archive file
    print('+ Adding certificate ' + pretty_filename)
    archived_cert_filepath = args.basedir + "/.archive/" + archived_cert_filename
    if os.path.exists(archived_cert_filepath):
        print("  This certificate is already archived")
        pass
    else:
        # Pretty print the certificate file
        x509pipe = pipes.Template()
        x509pipe.append('openssl x509 -text -nameopt oneline,-esc_msb,utf8', '--')
        cert_file = x509pipe.open(archived_cert_filepath, 'w')
        cert_file.write(crypto.dump_certificate(crypto.FILETYPE_PEM, certificate).decode("utf-8"))
        cert_file.close()

        if args.git:
            repo.index.add([archived_cert_filepath])

    # Set the file date to the beginning of the certificate's validity
    archived_cert_file_timestamps = time.mktime(
        time.strptime(certificate.get_notBefore().decode("utf-8"), '%Y%m%d%H%M%SZ')
    )
    os.utime(
        archived_cert_filepath,
        (
            archived_cert_file_timestamps,
            archived_cert_file_timestamps
        )
    )

    # Create a symlink from the archive file to ALL and Validity/Valid directories
    # and set the file dates accordingly
    archived_cert_rel_filepath = '.archive/' + archived_cert_filename

    # ALL directory: This link should be there all the time
    symlink_all = args.basedir + '/ALL/' + pretty_filename
    _link_file(archived_cert_rel_filepath, symlink_all)
    set_symlink_timestamps_to_target_timestamps(symlink_all, archived_cert_filepath)

    # Valid: If there is a corresponding file in the Validity subtree, we are good
    # Only create the symlink if this is not the case
    files = glob.glob(args.basedir + '/Validity/*/' + pretty_filename)
    if len(files) == 0:
        symlink_valid = args.basedir + '/Validity/Valid/' + pretty_filename
        _link_file(archived_cert_rel_filepath, symlink_valid)
        set_symlink_timestamps_to_target_timestamps(symlink_valid, archived_cert_filepath)

    # Append commit message
    global commits
    commits.append('Added ' + str(certificate.get_serial_number()))

    # Classify new certificate
    classify_by_serial(str(certificate.get_serial_number()))


# Read a certificate from a file and process it
# Takes file name as input
# If file name is '-', then stdin is read instead
def add_certificate_from_file(cert_file):
    if cert_file == '-':
        # Read the certificate from stdin
        cert_string = ''
        for line in stdin:
            cert_string += line
    else:
        # Read the certificate from file
        with open(cert_file, 'r') as fh:
            cert_string = fh.read()

    # Process the certificate
    add_certificate(cert_string)


# Read a certificate from a mail file and process it
def add_certificate_from_mail(mail):
    # Do some sanity checks
    subject = mail.get('Subject')
    if not subject or 'KIT-CA Zertifikatinformation' not in subject:
        return

    # Parse the email
    for part in mail.walk():
        # Do some sanity checks
        if part.get_content_type() != 'text/plain':
            continue
        if not part.get_filename():
            continue
        if not part.get_filename().startswith('cert-'):
            continue
        if not part.get_filename().endswith('.pem'):
            continue

        # Process the certificate
        add_certificate(part.get_payload(decode=True))


# Initialize directory structure
def init():
    """Initialize base directory structure for certificate archive"""
    global repo
    global commit_message

    try:
        _mkdir(args.basedir)
        if args.git:
            commit_message += 'Initializing: '
            repo = git.Repo.init(args.basedir)
            assert repo.bare is False

        for directory in [
                '.archive',
                'ALL',
                'Generation',
                'Generation/UNKNOWN',
                'Generation/KIT-CA-G1',
                'Generation/KIT-CA-G2',
                'Generation/geant-tcs-sectigo',
                'Profile',
                'Profile/802.1X_Client',
                'Profile/802.1X_User',
                'Profile/Code_Signing',
                'Profile/Domain_Controller',
                'Profile/Exchange_Server',
                'Profile/LDAP_Server',
                'Profile/Mail_Server',
                'Profile/Mitarbeiter',
                'Profile/OCSP_Responder',
                'Profile/Radius_Server',
                'Profile/RA_Operator',
                'Profile/Shibboleth_IdP_SP',
                'Profile/Smartcard',
                'Profile/Smartcard_Encrypt',
                'Profile/Smartcard_Logon',
                'Profile/Smartcard_Sign',
                'Profile/Smartcard_SignAndLogon',
                'Profile/Student',
                'Profile/UNKNOWN',
                'Profile/User',
                'Profile/UserAuth',
                'Profile/UserEMail',
                'Profile/UserEncrypt',
                'Profile/UserSign',
                'Profile/UserSignAndLogon',
                'Profile/UserSignAuth',
                'Profile/VoIP_Server',
                'Profile/VPN_Server',
                'Profile/VPN_User',
                'Profile/Web_Server',
                'Profile/Webserver_MustStaple',
                'Validity',
                'Validity/Expired',
                'Validity/Revoked',
                'Validity/UNKNOWN',
                'Validity/Valid',
                'Visibility',
                'Visibility/Private',
                'Visibility/Public',
                'Visibility/UNKNOWN',
        ]:
            _mkdir(args.basedir + '/' + directory)
            _link_file('../.archive',
                       args.basedir + '/' + directory + '/.archive')
            # Append commit message
            global commits
            commits.append('Created directory ' + directory)
    except OSError as e:
        print('Cannot initialize:', e)
        exit(1)


# Revoke certificates by serial number
def revoke_by_serial(serial):
    for directory in ['Validity/Expired', 'Validity/UNKNOWN', 'Validity/Valid']:
        for file in glob.glob(args.basedir + '/' + directory + '/' + '*-' + serial + '.pem'):
            print('! Revoking certificate ' + serial)
            _movefile(file, args.basedir + '/Validity/Revoked/')

            global commits
            commits.append('Revoked ' + serial)


# Revoke certificate by mail file
def revoke_by_mail(mail):
    # Do some sanity checks
    subject = mail.get('Subject')
    if not subject or 'DFN-PKI: Ihr Zertifikat ' not in subject or ' wurde gesperrt' not in subject.replace('\n', ''):
        return

    # Parse the email
    match = re.search(
        r'DFN-PKI: Ihr Zertifikat .*[\r\n]*.* \((\d+)\) wurde gesperrt',
        mail.get('Subject'),
        re.MULTILINE
    )
    if match:
        revoke_by_serial(match.group(1))


# Validate domain from URL
def validate_domain_from_url(domain, url):
    print('DV: Validating domain "', domain, '"...')
    resolver = dns.resolver.Resolver()
    resolver.nameservers = [socket.gethostbyname('dns1.kit.edu'), socket.gethostbyname('dns2.kit.edu')]
    try:
        resolver.query(domain, 'NS')
        print('DA: Domain "', domain, '" is known by KIT nameservers.  Approving...')
        response = urllib3.PoolManager().request('GET', url.replace('/domain/', '/domainconfirm/')).data
        if '' in response:
            print('DG: Domain "', domain, '" is approved with DFN-PKI.')
        else:
            print('DB: Error while approving domain "', domain, '"!')
    except dns.resolver.NoNameservers:
        print('DE: Domain "', domain, '" is NOT known by KIT nameservers.  NOT APPROVING.')


# Process mail file
def process_mail(mail):
    # Do some sanity checks
    subject = mail.get('Subject')

    # Fork by subject
    if not subject:
        print("[Error - processmail]: No subject given.")
    elif 'DFN-PKI: Ihr Zertifikat ' in subject and ' wurde gesperrt' in subject.replace('\n', ''):
        revoke_by_mail(mail)
    elif 'KIT-CA Zertifikatinformation' in subject:
        add_certificate_from_mail(mail)
    else:
        print("[Error - processmail] Mail is neither for a new certificate nor for a revocation.")


# Expire certificates by serial number
def expire_by_serial(serial):
    print('- Expiring certificate ' + serial)
    for directory in ['Validity/UNKNOWN', 'Validity/Valid']:
        for file in glob.glob(args.basedir + '/' + directory + '/' + '*-' + serial + '.pem'):
            _movefile(file, args.basedir + '/Validity/Expired/')
    global commits
    commits.append('Expired ' + serial)


# Classify certificates by serial number
def classify_by_serial(serial):
    commit_message_part = 'Classified ' + serial
    actions_done = []
    print('# Classifying certificate ' + serial)

    # Check whether the certificate is known
    archived_cert_filename = 'cert-' + serial + '.pem'
    archived_cert_rel_filepath = '.archive/' + archived_cert_filename
    archived_cert_filepath = args.basedir + "/" + archived_cert_rel_filepath

    if os.path.exists(archived_cert_filepath):
        files = glob.glob(args.basedir + '/ALL/*-' + serial + '.pem')
        if len(files) != 0:
            pretty_filename = os.path.basename(files[0])

            generation = "UNKNOWN"

            # File-content based check for CA generation
            files = glob.glob(args.basedir + '/Generation/*/' + pretty_filename)
            if len(files) == 0:
                with open(archived_cert_filepath, 'r') as file:
                    content = file.read()
                    for URL in CRL_URLS_TO_GENERATION.keys():
                        if URL in content:
                            generation = CRL_URLS_TO_GENERATION[URL]
                            break
                actions_done.append(f"Generation: {generation}")
                _link_file(
                    archived_cert_rel_filepath,
                    f"{args.basedir}/Generation/{generation}/{pretty_filename}"
                )

            # Callout-based check for visibility
            files = glob.glob(args.basedir + '/Visibility/*/' + pretty_filename)
            if args.callout and len(files) == 0:
                try:
                    response = urllib3.PoolManager().request('GET', 'https://www.ca.kit.edu/p/getcert/' + serial)
                    decoded_data = response.data.decode("utf-8")
                    if decoded_data.startswith('-----BEGIN CERTIFICATE-----'):
                        # Certificate is public
                        actions_done.append('Visibility: public')
                        _link_file(
                            archived_cert_rel_filepath,
                            args.basedir + '/Visibility/Public/' + pretty_filename
                        )
                    else:
                        # Certificate is private
                        actions_done.append('Visibility: private')
                        _link_file(archived_cert_rel_filepath, args.basedir + '/Visibility/Private/' + pretty_filename)
                except Exception as e:
                    send_notification(f"  Classification failed: {e}")
                    _link_file(archived_cert_rel_filepath, args.basedir + '/Visibility/UNKNOWN/' + pretty_filename)
                    actions_done.append('Visibility: UNKNOWN')

            # Heuristics-based approach to profiles
            # Currently: Everything is UNKNOWN
            files = glob.glob(args.basedir + '/Profile/*/' + pretty_filename)
            if len(files) == 0:
                actions_done.append('Profile: UNKNOWN')
                _link_file(archived_cert_rel_filepath, args.basedir + '/Profile/UNKNOWN/' + pretty_filename)

    global commits
    commits.append(commit_message_part + ' (' + ', '.join(actions_done) + ')')


# Wrapper functions
def add_by_cert_file_args():
    global commit_message
    commit_message += 'Add: '
    for argument in args.arguments:
        add_certificate_from_file(argument)


def add_by_mail_args():
    global commit_message
    commit_message += 'Addmail: '
    for argument in args.arguments:
        add_certificate_from_mail(read_mail(argument))


def revoke_by_serial_args():
    global commit_message
    commit_message += 'Revoke: '
    for argument in args.arguments:
        if argument == '-':
            for inputline in stdin:
                serial = inputline.rstrip()
                if serial:
                    revoke_by_serial(serial)
        else:
            revoke_by_serial(argument)


def revoke_by_mail_args():
    global commit_message
    commit_message += 'Revokemail: '
    for argument in args.arguments:
        revoke_by_mail(read_mail(argument))


def process_mail_args():
    global commit_message
    commit_message += 'Processmail: '
    for argument in args.arguments:
        process_mail(read_mail(argument))


def auto_expire():
    global commit_message
    commit_message += 'Autoexpire: '
    for directory in ['Validity/UNKNOWN', 'Validity/Valid']:
        for file in glob.glob(args.basedir + '/' + directory + '/' + '*.pem'):
            date = re.search(r'-(\d\d\d\d-\d\d-\d\d)-(\d+).pem$', file)
            if date and datetime.strptime(date.group(1) + 'T23:59', '%Y-%m-%dT%H:%M') < datetime.now():
                expire_by_serial(date.group(2))


def auto_revoke():
    global commit_message
    commit_message += 'Autorevoke: '
    if args.callout:
        for crl in CRL_URLS_TO_GENERATION.keys():
            try:
                revoked_certs = crypto.load_crl(
                    crypto.FILETYPE_ASN1, urllib3.PoolManager().request('GET', crl).data
                ).get_revoked()
                if revoked_certs:
                    print(f"Checking {len(revoked_certs)} revoked certificates from {crl}...")
                    for idx, revocation in enumerate(revoked_certs):
                        revoke_by_serial(str(int(revocation.get_serial(), 16)))
            except:
                message = 'Could not load CRL ' + crl
                commit_message += message + '; '
                print(message)


def expire_by_serial_args():
    global commit_message
    commit_message += 'Expire: '
    for argument in args.arguments:
        if argument == '-':
            for inputline in stdin:
                serial = inputline.rstrip()
                if serial:
                    expire_by_serial(serial)
        else:
            expire_by_serial(argument)


def classify_by_serial_args():
    global commit_message
    commit_message += 'Classify: '
    for argument in args.arguments:
        if argument == '-':
            for inputline in stdin:
                serial = inputline.rstrip()
                if serial:
                    classify_by_serial(serial)
        else:
            classify_by_serial(argument)


def send_notification(message: str):
    if not args.notification:
        # Print instead
        print("[Notification] " + message)
        return

    if not notifier_config:
        return

    try:
        urllib3.PoolManager().request_encode_body(
            "POST",
            SIGNAL_URL + notifier_config["signal_target"],
            headers={"Authorization": f"Bearer {notifier_config['signal_token']}"},
            fields={"text": message}
        )
    except Exception as e:
        print(f"WARNING: Failed to send notification - {e}")


def load_notifier_config():
    global notifier_config

    try:
        with open("/var/ca-procmail/.config/process_certificate.json", "r") as fh:
            try:
                notifier_config = json.load(fh)
            except json.JSONDecodeError as e:
                print(f"WARNING: Failed to load notifier config - {e}")

            if not all([notifier_config.get(key) for key in ["signal_target", "signal_token"]]):
                print("WARNING: Notifier config incorrect, requires key signal_target and signal_token")

    except (FileNotFoundError, IOError) as e:
        print(f"WARNING: Notifier config could not be opened - {e}")


def dispatch(cli_args):
    global repo
    global commit_message

    # Initialize a new git repository if requested
    if cli_args.action == 'init':
        init()
        return

    if cli_args.git:
        repo = git.Repo(cli_args.basedir)
        assert repo.bare is False

    if cli_args.action == 'add':
        add_by_cert_file_args()
    elif cli_args.action == 'addmail':
        add_by_mail_args()
    elif cli_args.action == 'classify':
        classify_by_serial_args()
    elif cli_args.action == 'expire':
        expire_by_serial_args()
    elif cli_args.action == 'revoke':
        revoke_by_serial_args()
    elif cli_args.action == 'revokemail':
        revoke_by_mail_args()
    elif cli_args.action == 'processmail':
        process_mail_args()
    elif cli_args.action == 'autoexpire':
        auto_expire()
    elif cli_args.action == 'autorevoke':
        auto_revoke()
    elif cli_args.action == 'autoprocess':
        commit_message = 'Autoprocess: '
        auto_expire()
        auto_revoke()
    elif cli_args.action == 'testnotification':
        send_notification("This is a test notification from process_certificate.py")
        return

    # Finally, commit changes
    if cli_args.git and repo.is_dirty():
        repo.index.commit(commit_message + ', '.join(commits))
        if cli_args.gitremotes:
            for remote in repo.remotes:
                remote.push()
        if os.path.isfile(cli_args.basedir + '/.git/hooks/post-commit'):
            subprocess.call(cli_args.basedir + '/.git/hooks/post-commit', cwd=cli_args.basedir)


def dispatch_wrapper():
    try:
        dispatch(args)
    except Exception as e:
        # Something broke somewhere, send a notification and exit non-successfully

        message = "\n".join([f"    {line}" for line in str(e).split("\n")])
        args_str = "\n".join([f"    {key}: {value}" for key, value in vars(args).items()])
        text = (
            f"Error in process_certificate:\n\n"
            f"Type: {type(e).__name__}\n"
            f"Message:\n{message}\n\n"
            f"Called with arguments:\n{args_str}\n\n"
        )
        send_notification(text + "Printed stacktrace to stdout. Please check logs.")
        print(f"{datetime.now()}\n{text}{traceback.format_exc()}")
        exit(1)


def main():
    global args

    parser = ArgumentParser(description='Handles certicificate archive')
    parser.add_argument('--basedir', dest='basedir', default=DEFAULT_BASEDIR, type=str, help='archive base directory')
    parser.add_argument('--no-callout', dest='callout', action='store_false', default=True)
    parser.add_argument('--no-git', dest='git', action='store_false', default=True)
    parser.add_argument('--no-remotes', dest='gitremotes', action='store_false', default=True)
    parser.add_argument('--no-notification', dest='notification', action='store_false', default=True,
                        help="Print notification to stdout instead of sending it via signal.")
    parser.add_argument('action', choices=ACTIONS, type=str, help='operation to be performed')
    parser.add_argument('arguments', nargs='*', help='arguments to the operation')
    args = parser.parse_args()

    # Make basedir canonical
    args.basedir = os.path.realpath(args.basedir)

    load_notifier_config()

    lock = filelock.FileLock(LOCK_FILE)
    try:
        with lock.acquire(timeout=LOCK_TIMEOUT):
            dispatch_wrapper()
    except filelock.Timeout:
        send_notification("Failed to acquire lock for process_certificate!")


if __name__ == "__main__":
    main()

